# @auto-fold regex /^\s*if/ /^\s*else/ /^\s*def/
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Metadata about the package
__version__ = '0.9'
__name__ = 'astroemperor'

from .emp import Simulation

__all__ = ['Simulation']